extends Node3D

@export var tracking : UltraleapHandTracking
@export var materials_to_rectify : Array[Material]

@export_group("Layers")
@export var disconnected : CanvasLayer
@export var no_device : CanvasLayer

signal view_size_changed(view_size : Vector2i)
signal current_device_tracking_mode_changed(tracking_mode : int)
signal presence_of_device_changed(is_present : bool)
signal current_device_changed(device : UltraleapDevice)

var current_device : UltraleapDevice

var previous_tracking_mode : int


# Called every frame. 'delta' is the elapsed time since the previous frame.
func _process(_delta):
	if current_device == null:
		return
	
	if current_device.get_tracking_mode() != previous_tracking_mode:
		emit_signal("current_device_tracking_mode_changed", current_device.get_tracking_mode())
		
	previous_tracking_mode = current_device.get_tracking_mode()


func _on_device_selection(serial):
#	if current_device != null:
#		_close_device()
#	current_device = _get_device_from_serial(serial)
#	_open_device()
#	current_device_changed.emit(current_device)
	var dev : UltraleapDevice = _get_device_from_serial(serial)
	var current_dev : UltraleapDevice = tracking.get_primary_device()
	
	if current_dev == null or current_dev.id != dev.id:
		tracking.set_primary_device(dev)
		emit_signal("current_device_changed", dev)


func _get_device_from_serial(serial : String):
	for device in tracking.devices:
		if device.serial == serial:
			return device


func _open_device():
	current_device.open()
	current_device.subscribe()


func _close_device():
	current_device.unsubscribe()
	current_device.close()


func _on_hand_tracking_device_added(device):
	if tracking.devices.size() > 0:
		emit_signal("presence_of_device_changed", true)
		no_device.hide()
#	if current_device == null:
#		current_device = device
#		current_device_changed.emit(current_device)
#		_open_device()


func _on_hand_tracking_device_removed(_device):
	if tracking.devices.size() == 0:
		emit_signal("presence_of_device_changed", false)
		no_device.show()
	if current_device == null:
		return
	_close_device()
	current_device = null
	current_device_changed.emit(null)


func on_undistort_toggled(enabled):
	for material in materials_to_rectify:
		material.set_shader_parameter("undistort", enabled)


func _on_connection_status_changed(connected):
	if connected:
		emit_signal("presence_of_device_changed", false)
		no_device.show()
		disconnected.hide()
	else:
		emit_signal("presence_of_device_changed", false)
		no_device.hide()
		disconnected.show()
